#ifndef _ELF_TYPES_H
#define _ELF_TYPES_H

// #include "ELFConst.h"
typedef uint16_t eeprompage_t;
typedef uint16_t eeprompageoffset_t; /* 0 to TOS_EEPROM_PAGE_SIZE - 1 */

typedef uint16_t flashpage_t;
typedef uint16_t flashpageoffset_t; /* 0 to TOS_EEPROM_PAGE_SIZE - 1 */


typedef eeprompage_t fileblock_t;
typedef eeprompageoffset_t fileblockoffset_t; /* 0 to IFS_PAGE_SIZE */

// typedef eeprompage_t flashpage_t;
typedef eeprompageoffset_t filepageoffset_t; /* 0 to IFS_PAGE_SIZE */

typedef uint8_t fileblockfoo_t; /* 0 to IFS_PAGE_SIZE - 1 */
typedef uint32_t filemeta_t;	/* Metadata version number */

typedef uint8_t fmode_t;

typedef uint8_t fileresult_t;
typedef uint32_t filesize_t;

enum {
  TOS_EEPROM_MAX_PAGES = 2048,
  TOS_EEPROM_PAGE_SIZE = 264,
  TOS_EEPROM_PAGE_SIZE_LOG2 = 8 // For those who want to ignore the last 8 bytes
};

enum {
  TOS_EEPROM_ERASE,
  TOS_EEPROM_DONT_ERASE,
  TOS_EEPROM_PREVIOUSLY_ERASED
};


enum {
  FS_OK,
  FS_NO_MORE_FILES,
  FS_ERROR_NOSPACE,
  FS_ERROR_BAD_DATA,
  FS_ERROR_FILE_OPEN,
  FS_ERROR_NOT_FOUND,
  FS_ERROR_HW
};

enum {
  FS_FTRUNCATE = 1,
  FS_FCREATE = 2
};

enum{
    PAGE_DIRTY = 0x08,
    PAGE_END = 0x04,
    PAGE_FREE = 0x02,
    PAGE_INUSE = 0x01
};

enum {
  IFS_PAGE_SIZE = 256,
  IFS_LOG2_PAGE_SIZE = 8,
  IFS_FULL_PAGE_SIZE = 264,

  IFS_FIRST_PAGE = 0,
  IFS_NUM_PAGES = 1536
};

enum {
  IFS_OFFSET_METADATA = 258,
  IFS_EOF_BLOCK = IFS_NUM_PAGES + 42,
  IFS_ROOT_MARKER = 4,
  IFS_ROOT_MARKER_BITS = 3
};

enum {
  IFS_RFD_META = unique("FileRead"),
  IFS_WFD_META = unique("FileWrite")
};

// note: the original name is struct page_info instead of metadata
struct metadata{
  uint16_t crc;
  uint16_t nextPage:11;  // The lower 11 bit , point to the next page in inode
  uint16_t flag:5;       // The 5 MSB, they are dirty/free/in_used/inited/
  uint16_t writeCounter; // 
  uint16_t magic;        // magic number to identify the inode 
} __attribute__((packed));
typedef struct metadata flashpageinfo_t;


/*     ---    Data Structures Representation on Flash --- */

/* This is the inode structure for the first physical node of the file
 * It stores most of the file information.  Currently we image
 * there are few renaming action and the name size is not too long
 * 17 bytes.  It's followed by the name of the file. 
 * Any modification to this node results in a copy of the content to
 * the new page
 */
struct elf_major_inode
{
  uint8_t nodeType;       // The type of inode, major inode, common, and directory   1
  uint16_t version;       // Version number of this physical node.  should be less than 10,000 times  2 bytes
  uint16_t inodeNumber ;  // The inode No. of the file/dir this physical belongs to 2 bytes
  uint32_t locaLen;       // length of this physical node, in byte 4( can be 1 byte smaller using page# + last byte)
  uint16_t pInode;        // the parent inode, 0 for top level
  uint8_t spare : 3;      // For future use.                           
  uint8_t reliable : 2;   // Should this file be treated as a reliable file
  uint8_t fragmented : 1;    // Has this file been modified?        1
  uint8_t rename : 1;     // Has this file been renamed ?
  uint8_t deleted : 1;    // Has this file been deleted?        1
  uint8_t  nsize;         // Name length.                  1
  uint32_t totaLen;       // The total length of the file represted by this inode
} __attribute__((packed));
typedef struct elf_major_inode elf_major_node_t;


/* This is the inode structure for the other nodes in the file.
 * The file might be segmented on the flash memory
 * 14 bytes
 */
struct elf_minor_inode
{
  uint8_t nodeType;    // The type of inode, major inode, common, and directory  1
  uint16_t version;    // Version number.  should be less than 10,000 times  2 bytes
  uint16_t inodeNumber ;   // The inode No. of the file/dir this physical belongs to  2 byte
  uint32_t locaLen;    // The total length of this file, in byte.     4
  uint8_t spare : 3;      // For future use.                           
  uint8_t localId : 4;      // for comparison in the EEPROM.                           
  uint8_t deleted : 1;    // Has this node been deleted?        1
  uint32_t  beginOffset;        // where to begin to write           4
} __attribute__((packed));
typedef struct elf_minor_inode elf_minor_node_t;


/* This is the inode structure for the directory, should I pay extra 
   attention to it? since directory inode could be pretty small 
   potentially
*/
// In the flash, after the last field, the real directory data begins.
// First is the directory name, which is nsize in length.  It is
// followed by the inodeNumbers in the flash memory.
struct elf_dirent{
  uint8_t nodeType;    // The type of inode  1
  uint16_t version;       // Version number.   2
  uint16_t inodeNumber;      // The inode number of this directory 2
  uint32_t locaLen;      // The total size of this dir inode 4
  uint16_t pInode;        // the parent inode, 0 for top level 2
  uint8_t spare : 3;      // For future use.                           
  uint8_t reliable : 2;   // Should this file be treated as a reliable file
  uint8_t modified : 1;    // Has this file been modified?        
  uint8_t rename : 1;     // Has this file been renamed ?
  uint8_t deleted : 1;    // Has this file been deleted?        1
  uint8_t nsize;            // length of name 1
  uint8_t childInodeNum;  // The number of children files/directories 1
  uint32_t crc;            // 4 bytes for this metadata
} __attribute__((packed));
typedef struct elf_dirent elf_dir_t;




/*     ---    Node Structures Representation in Memory  --- */

/* Reference to an inode on the disk */
#define MAX_INODE_SIZE 15
struct elf_raw_node
{
  uint8_t nodeType;    // Inode number.  
  uint16_t version;        // Version number.  
  uint16_t inodeNumber;      // The inode number of this directory 2
  uint32_t locaLen;      // The total size of this dir inode 4
  uint8_t  contents[MAX_INODE_SIZE];
} __attribute__((packed));
typedef struct elf_raw_node elf_raw_node_t;



/* The RAM representation of the node. It is the logic representation
   of each of the inode in the system.  To save memory, we only 
   keep a small number of informations.  The minimum number results in the 
   operation inefficiency, what should I do
   15 bytes
 */
struct elf_node
{
  uint32_t locaLen;              //  4 bytes
  uint32_t beginOffset;          //  4 bytes
  uint16_t version;              //  Version number of this inode 2 byte
  uint16_t startPg : 11;         //  Physical page of the data in the actual 2 byte
  uint16_t flag: 5;              //  flag
  uint8_t  localId;              // use to identify which node to be replaced  1 byte
  struct elf_node *next;         // next node belongs to the same inode 2 byte
} __attribute__((packed));
typedef struct elf_node elf_node_t;

/* For each inode in the filesystem, we keep a cache that indicates
   the begining of the node. We also keep a pointer to the first
   physical node which is part of this inode, too.

   need to change to fit the directory structure
*/

// add cache supported , 20 bytes in total

struct elf_inode {
  uint8_t nodeType;  //1
  uint16_t version;  //2 
  uint16_t inodeNumber;          // the inode number of this node 2
  uint32_t totaLen;              // total length of this file 4
  uint16_t pInode;               // the parent inode, 0 for top level  2
  uint8_t flags;                 // locked|opened|rename|deleted|reliable 1
  uint8_t nsize;  // 1
  uint8_t readCount : 4 ;
  uint8_t writeCount : 4 ;       // thread # that is currently working on the system 1
  struct elf_inode *next;         // the next inode link in the same level 2 byte
  elf_node_t *first;             // the first physical node  2 byte
  //  well, here should have a lock to guarantee the access to the file
  uint8_t *cache;  // 2
} __attribute__((packed));
typedef struct elf_inode elf_inode_cache_t;
elf_inode_cache_t root;


struct elf_page_cache{
  uint16_t page;
  uint16_t age; 
}__attribute__((packed));
typedef struct elf_page_cache elf_page_cache_t;

#define PAGE_CACHES       10
struct elf_resources_bitmap{
  uint8_t  bitmap[256];                 // the bitmap of the flash
  //  uint32_t totalAge;                      // whether to allocate a page
  uint16_t freePgNum;                   // save the time on scanning the bitmap
  uint16_t gc_minfree_threshold;	// GC trigger thresholds 
  uint16_t gc_maxdirty_threshold;       // allowed to be adjusted
  elf_page_cache_t youngPgs[PAGE_CACHES];
}__attribute__((packed));

// currently use the bitmap, to be changed in the future
typedef struct elf_resources_bitmap elf_resource_t;
elf_resource_t res;


/* Resource Management Approach 2 : free list/dirty block

   During the garbage colleciton, the list with free blocks are
   built.  When a file is to be erased, its pages are load into 
   the buffer first, and check whether it can be merged with the 
   existing free lists.  These work are done in the garbage 
   collection that executes once a minute?
*/
struct elf_resources_list{
  uint16_t beginPage;
  uint16_t pageNum;
  struct elf_resources_list* next;
}__attribute__((packed));


struct file_descriptor{
  fmode_t flags;            // 1 byte 
  elf_inode_cache_t *file;  // the file this descriptor points to 2 byte
  uint32_t curOffset;       // offset in the file    4 byte;
  elf_node_t *curNode;      // 2 bytes  
  uint16_t lastPg;          // 2 byte
  uint8_t lastByte;         // 1 byte
  //		remaining reserved byte for write descriptor?
} __attribute__((packed));
typedef struct file_descriptor elf_fd_t;


struct file_offset{
  elf_node_t *node;  
  uint16_t page;
  uint8_t inPgOffset;
  uint32_t offset;
} __attribute__((packed));
typedef struct file_offset elf_file_offset_t;

#define MAX_OPENED_FILES 3
elf_fd_t fds[MAX_OPENED_FILES];

// elf_fm.h

struct elf_control
{
  elf_inode_cache_t *root;	      // The root directory file.  
  //  struct list_head *hash;	      // Hash table for finding files by ino. 
  //  uint16_t hash_len;	// The size of the hash table.  
  uint16_t nextIno;		// Next inode number to use for new files.  
  uint16_t indexPg;             // index page containing the name-inode pair
  elf_resource_t *elfResource;
 } __attribute__((packed));
typedef struct elf_control elf_control_t;
elf_control_t control, *elfControl; // the only global structure


struct elf_index_file_list{
  uint8_t nameLen;
  uint16_t inode;
  uint16_t offset;
} __attribute__((packed));
typedef struct elf_index_file_list index_filelist_t;

#define MAX_FILES 10
// This is at the top of the index page
struct elf_index_summary{
  uint8_t count; 
  index_filelist_t fileList[MAX_FILES];
  uint16_t endPtr;
  uint16_t checksum;
}__attribute__((packed));
typedef struct elf_index_summary index_summary_t;




// This structure starts at the SUMMARY_OFFSET position
// 16 byte + resource size (262 byte)
struct elf_eeprom_summary
{
  uint16_t magic;   // elf magic number to identify, use ELF_GENERAL 2
  uint8_t flag;     // 1
  uint8_t version; // elf version number 1
  uint8_t fileCount;  //  1 
  uint8_t nodeCount;  // 1
  uint16_t snapshotSummaryOffset; // 2
  uint16_t metaDataOffset;        // 2
  uint16_t snapShotOffset;        // 2
  elf_resource_t resource;  // resource abstraction in ELF
  uint16_t pgNameIndex;    // 2
  uint16_t checksum;       // 2
} __attribute__((packed));
typedef struct elf_eeprom_summary elf_eeprom_summary_t;


struct elf_snapshot_summary
{
  uint8_t  count;
  uint8_t  flag;
  uint16_t inodes[MAX_OPENED_FILES];
  uint16_t checksum;
}__attribute__((packed));
typedef struct elf_snapshot_summary elf_snapshot_summary_t;

// 6 bytes
struct elf_oper_summary
{
  uint16_t inode;
  uint8_t  operCount;
  uint8_t  snapshotTail;
  uint16_t checksum;
}__attribute__((packed));
typedef struct elf_oper_summary elf_oper_summary_t;

struct elf_metadata_summary
{
  uint16_t inode;
  uint8_t  flag;  
}__attribute__((packed));
typedef struct elf_metadata_summary elf_metadata_summary_t;


// 13 bytes in total
struct elf_inode_snapshot {
  uint8_t nodeType;          // 1
  uint16_t version;          // 2
  uint16_t inode;          // the inode number of this node 2
  uint32_t totaLen;              // total length of this file 4
  uint16_t pInode;               // the parent inode, 0 for top level  2
  uint8_t flags;                 // last bit indicate whether it's a dir   |rename|deleted 1
  uint8_t nsize;         // 1
} __attribute__((packed));
typedef struct elf_inode_snapshot elf_inode_snapshot_t;


#define OPER_DATA_BUFFER_LEN 12
struct elf_operation_snapshot
{
  uint8_t oper;
  uint16_t inode;
  uint8_t operData[OPER_DATA_BUFFER_LEN];
}__attribute__((packed));
typedef struct elf_operation_snapshot elf_snapshot_t;

/*
  Operation Snapshot Formats:
*/
  // class1 operation :  write, append 
  struct elf_snapshot_oper1
  {
    uint8_t oper;
    uint16_t inode;
    uint16_t startPage;
    uint32_t startOffset;
    uint32_t size;
    uint16_t version;
  }__attribute__((packed));
typedef struct elf_snapshot_oper1 elf_snapshot_oper1_t;

  // class2 operation : delete, create, rename
  struct elf_snapshot_oper2
  {
    uint8_t oper;
    uint16_t inode;
    uint16_t page;
    uint16_t version;
  }__attribute__((packed));
typedef struct elf_snapshot_oper2 elf_snapshot_oper2_t;


#endif
